import json

from setup_tests import Path, run_request, test_dir, write_rpc_request


def hover_req(file_path: Path, ln: int, col: int) -> str:
    return write_rpc_request(
        1,
        "textDocument/hover",
        {
            "textDocument": {"uri": str(file_path)},
            "position": {"line": ln, "character": col},
        },
    )


def validate_hover(result_array: list, checks: list):
    assert len(result_array) - 1 == len(checks)
    for i, check in enumerate(checks):
        assert result_array[i + 1]["contents"]["value"] == check


def test_hover_abstract_int_procedure():
    """Tests that the binding of an abstract interface is correctly resolved"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_abstract.f90"
    string += hover_req(file_path, 7, 30)
    errcode, results = run_request(string, fortls_args=["--sort_keywords", "-n1"])
    assert errcode == 0
    ref_results = [
        """```fortran90
SUBROUTINE test(a, b)
 INTEGER(4), DIMENSION(3,6), INTENT(IN) :: a
 REAL(8), DIMENSION(4), INTENT(OUT) :: b
```"""
    ]
    validate_hover(results, ref_results)


def test_hover_parameter_multiline():
    """Test that hover parameters display value correctly across lines"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 2, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nINTEGER, PARAMETER :: var = 1000\n```"]
    validate_hover(results, ref_results)


def test_hover_literal_num():
    """Test that hovering over literals shows their type INTEGER"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 3, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nINTEGER\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter():
    """Test that hover parameters display value correctly"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 4, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nINTEGER, PARAMETER :: var2 = 23\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_dollar():
    """Test that hover parameters with dollar in name are recognized correctly"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 20, 31)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nINTEGER(4), PARAMETER :: SIG$ERR = -1\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_eqnospace():
    """Test that hover parameters display value correctly"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 11, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nINTEGER, PARAMETER :: var_no_space = 123\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_morespace():
    """Test that hover parameters display value correctly"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 12, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nINTEGER, PARAMETER :: var_more_space = 123\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_var_sum():
    """Test that hover parameters display value correctly with sum"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 13, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nINTEGER, PARAMETER :: var_sum1 = 1 + 23\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_var_neg():
    """Test that hover parameters display value correctly with extraction"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 14, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nINTEGER, PARAMETER :: var_ex1 = 1 - 23\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_var_mul():
    """Test that hover parameters display value correctly with
    multiplication and spaces"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 15, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nINTEGER, PARAMETER :: var_mul1 = 1 * 23\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_var_div():
    """Test that hover parameters display value correctly with value of division"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 16, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nINTEGER, PARAMETER :: var_div1 = 1/1\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_var_multiline2():
    """Test that hover parameters display value correctly with
    multiplication and spaces. Item 2"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 17, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = [
        "```fortran90\nINTEGER, PARAMETER :: var_multi2 = 1 * 23 + 2 /1\n```"
    ]
    validate_hover(results, ref_results)


def test_hover_parameter_nested():
    """Test that hover parameters using other parameter values works"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 4, 41)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nINTEGER, PARAMETER :: var3 = var*var2\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_multiline_missing_type():
    """Test that hover parameters display correctly when type is split across lines"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 6, 28)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nINTEGER, PARAMETER :: var4 = 123\n```"]
    validate_hover(results, ref_results)


def test_hover_literal_real():
    """Test that hovering over literals shows their values REAL"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 7, 47)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nREAL\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_double():
    """Test that hovering over parameters shows their type DOUBLE PRECISION"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 7, 38)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nDOUBLE PRECISION, PARAMETER :: somevar = 23.12\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_double_sf():
    """Test that hovering over parameters shows their type scientific notation"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 7, 55)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nDOUBLE PRECISION, PARAMETER :: some = 1e-19\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_bool():
    """Test that hovering over parameters shows their values LOGICAL"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 8, 38)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = [
        "```fortran90\nLOGICAL(kind=8), PARAMETER :: long_bool = .true.\n```"
    ]
    validate_hover(results, ref_results)


def test_hover_literal_bool():
    """Test that hovering over literals shows their type LOGICAL"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 8, 50)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nLOGICAL\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_str_sq():
    """Test that hovering over parameters shows their value, single quote STRING"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 9, 37)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nCHARACTER(len=5), PARAMETER :: sq_str = '12345'\n```"]
    validate_hover(results, ref_results)


def test_hover_literal_string_sq():
    """Test that hovering over literals shows their values single quote STRING"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 9, 48)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nCHARACTER(LEN=5)\n```"]
    validate_hover(results, ref_results)


def test_hover_parameter_str_dq():
    """Test that hovering over parameters shows their value, double quote STRING"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 10, 37)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ['```fortran90\nCHARACTER(len=5), PARAMETER :: dq_str = "12345"\n```']
    validate_hover(results, ref_results)


def test_hover_literal_string_dq():
    """Test that hovering over literals shows their values double quote STRING"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "parameters.f90"
    string += hover_req(file_path, 10, 48)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nCHARACTER(LEN=5)\n```"]
    validate_hover(results, ref_results)


def test_hover_pointer_attr():
    """Test that hovering maintains the variable attributes e.g. POINTER"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "pointers.f90"
    string += hover_req(file_path, 1, 26)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = ["```fortran90\nINTEGER, POINTER :: val1\n```"]
    validate_hover(results, ref_results)


def test_hover_functions():
    """Test that hovering over functions provides the expected results"""
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "functions.f90"
    string += hover_req(file_path, 1, 11)
    string += hover_req(file_path, 7, 19)
    string += hover_req(file_path, 12, 12)
    string += hover_req(file_path, 18, 19)
    string += hover_req(file_path, 23, 34)
    string += hover_req(file_path, 28, 11)
    string += hover_req(file_path, 34, 21)
    string += hover_req(file_path, 46, 11)
    string += hover_req(file_path, 51, 11)
    string += hover_req(file_path, 55, 11)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0

    ref_results = [
        """```fortran90
FUNCTION fun1(arg) RESULT(fun1)
 INTEGER, INTENT(IN) :: arg
 INTEGER :: fun1
```""",
        """```fortran90
FUNCTION fun2(arg) RESULT(fun2)
 INTEGER, INTENT(IN) :: arg
 INTEGER :: fun2
```""",
        """```fortran90
FUNCTION fun3(arg) RESULT(retval)
 INTEGER, INTENT(IN) :: arg
 INTEGER :: retval
```""",
        """```fortran90
FUNCTION fun4(arg) RESULT(retval)
 INTEGER, INTENT(IN) :: arg
 INTEGER :: retval
```""",
        # Notice that the order of the modifiers does not match the source code
        # This is part of the test, ideally they would be identical but previously
        # any modifiers before the type would be discarded
        """```fortran90
PURE ELEMENTAL FUNCTION fun5(arg) RESULT(retval)
 INTEGER, INTENT(IN) :: arg
 INTEGER :: retval
```""",
        """```fortran90
FUNCTION fun6(arg) RESULT(retval)
 INTEGER, INTENT(IN) :: arg
 INTEGER, DIMENSION(10,10) :: retval
```""",
        """```fortran90
PURE FUNCTION outer_product(x, y) RESULT(outer_product)
 REAL, DIMENSION(:), INTENT(IN) :: x
 REAL, DIMENSION(:), INTENT(IN) :: y
 REAL, DIMENSION(SIZE(X), SIZE(Y)) :: outer_product
```""",
        """```fortran90
FUNCTION dlamch(cmach) RESULT(dlamch)
 CHARACTER :: CMACH
```""",
        """```fortran90
FUNCTION fun7() RESULT(val)
 TYPE(c_ptr) :: val
```""",
        """```fortran90
TYPE(c_ptr) FUNCTION c_loc(x) RESULT(c_loc)
```""",
    ]
    validate_hover(results, ref_results)


def test_hover_spaced_keywords():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "spaced_keywords.f90"
    string += hover_req(file_path, 1, 45)
    string += hover_req(file_path, 2, 99)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = [
        """```fortran90\nREAL, DIMENSION(:, :), INTENT(IN) :: arg1\n```""",
        """```fortran90\nREAL, DIMENSION( SIZE(ARG1, 1), MAXVAL([SIZE(ARG1, 2), """
        """SIZE(ARG1, 1)]) ), INTENT(OUT) :: arg2\n```""",
    ]
    validate_hover(results, ref_results)


def test_hover_recursive():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "recursive.f90"
    string += hover_req(file_path, 9, 40)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = [
        """```fortran90
RECURSIVE SUBROUTINE recursive_assign_descending(node, vector, current_loc)
 TYPE(tree_inode), POINTER, INTENT(IN) :: node
 INTEGER, DIMENSION(:), INTENT(INOUT) :: vector
 INTEGER, INTENT(INOUT) :: current_loc
```"""
    ]
    validate_hover(results, ref_results)


def test_hover_subroutine():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "subdir" / "test_submod.F90"
    string += hover_req(file_path, 29, 24)
    string += hover_req(file_path, 34, 24)
    errcode, results = run_request(string, fortls_args=["--sort_keywords"])
    assert errcode == 0
    ref_results = [
        """```fortran90
FUNCTION point_dist(a, b) RESULT(distance)
 TYPE(point), INTENT(IN) :: a
 TYPE(point), INTENT(IN) :: b
 REAL :: distance
```""",
        """```fortran90
FUNCTION is_point_equal_a(a, b) RESULT(is_point_equal_a)
 TYPE(point), INTENT(IN) :: a
 TYPE(point), INTENT(IN) :: b
 LOGICAL :: is_point_equal_a
```""",
    ]
    validate_hover(results, ref_results)


def test_hover_interface_as_argument():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "test_diagnostic_int.f90"
    string += hover_req(file_path, 19, 14)
    errcode, results = run_request(string, fortls_args=["--sort_keywords", "-n1"])
    assert errcode == 0
    ref_results = [
        # Could be subject to change
        """```fortran90
FUNCTION foo2(f, g, h) RESULT(arg3)
 FUNCTION f(x)
 FUNCTION g(x)
 FUNCTION h(x)
 REAL :: arg3
```""",
    ]
    validate_hover(results, ref_results)


def test_hover_block():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "hover")})
    file_path = test_dir / "hover" / "associate_block.f90"
    string += hover_req(file_path, 4, 17)
    string += hover_req(file_path, 4, 20)
    # string += hover_req(file_path, 10, 11)    # slice of array
    errcode, results = run_request(string, fortls_args=["--sort_keywords", "-n", "1"])
    assert errcode == 0
    ref_results = [
        "```fortran90\nREAL, DIMENSION(5) :: X\n```",
        "```fortran90\nREAL :: Y\n```",
    ]
    validate_hover(results, ref_results)


def test_associate_block_func_result():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "hover")})
    file_path = test_dir / "hover" / "associate_block_2.f90"
    string += hover_req(file_path, 2, 14)
    string += hover_req(file_path, 3, 9)
    errorcode, results = run_request(string, fortls_args=["--sort_keywords", "-n", "1"])
    assert errorcode == 0
    ref_results = [
        "```fortran90\nLOGICAL FUNCTION :: hi\n```",
        "```fortran90\nLOGICAL FUNCTION :: hi\n```",
    ]
    validate_hover(results, ref_results)


def test_hover_submodule_procedure():
    """Test that submodule procedures and functions with modifier keywords
    are correctly displayed when hovering.
    """
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "diag")})
    file_path = test_dir / "diag" / "test_scope_overreach.f90"
    string += hover_req(file_path, 18, 37)
    string += hover_req(file_path, 23, 37)
    errcode, results = run_request(string, fortls_args=["-n", "1"])
    assert errcode == 0
    ref_results = [
        """```fortran90
PURE RECURSIVE FUNCTION foo_sp(x) RESULT(fi)
 REAL(sp), INTENT(IN) :: x
 REAL(sp) :: fi
```""",
        """```fortran90
PURE RECURSIVE FUNCTION foo_dp(x) RESULT(fi)
 REAL(dp), INTENT(IN) :: x
 REAL(dp) :: fi
```""",
    ]
    validate_hover(results, ref_results)


def test_var_type_kinds():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "parse")})
    file_path = test_dir / "parse" / "test_kinds_and_dims.f90"
    string += hover_req(file_path, 2, 24)
    string += hover_req(file_path, 2, 27)
    string += hover_req(file_path, 3, 15)
    string += hover_req(file_path, 3, 19)
    string += hover_req(file_path, 4, 20)
    string += hover_req(file_path, 4, 25)
    string += hover_req(file_path, 5, 23)
    string += hover_req(file_path, 6, 25)
    errcode, results = run_request(string, fortls_args=["-n", "1"])
    assert errcode == 0
    ref_results = [
        "```fortran90\nINTEGER(kind=4) :: a\n```",
        "```fortran90\nINTEGER(kind=4), DIMENSION(3,4) :: b\n```",
        "```fortran90\nINTEGER*8 :: aa\n```",
        "```fortran90\nINTEGER*8, DIMENSION(3,4) :: bb\n```",
        "```fortran90\nINTEGER(8) :: aaa\n```",
        "```fortran90\nINTEGER(8), DIMENSION(3,4) :: bbb\n```",
        "```fortran90\nREAL(kind=r15) :: r\n```",
        "```fortran90\nREAL(kind(0.d0)) :: rr\n```",
    ]
    validate_hover(results, ref_results)


def test_kind_function_result():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "parse")})
    file_path = test_dir / "parse" / "test_kinds_and_dims.f90"
    string += hover_req(file_path, 9, 18)
    string += hover_req(file_path, 14, 25)
    errcode, results = run_request(string, fortls_args=["-n", "1"])
    assert errcode == 0
    ref_results = [
        """```fortran90
FUNCTION foo(val) RESULT(r)
 REAL(8), INTENT(IN) :: val
 REAL*8 :: r
```""",
        """```fortran90
FUNCTION phi(val) RESULT(r)
 REAL(8), INTENT(IN) :: val
 REAL(kind=8) :: r
```""",
    ]
    validate_hover(results, ref_results)


def test_var_type_asterisk():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "parse")})
    file_path = test_dir / "parse" / "test_kinds_and_dims.f90"
    string += hover_req(file_path, 2 + 19, 18)
    string += hover_req(file_path, 2 + 19, 21)
    string += hover_req(file_path, 2 + 19, 29)
    string += hover_req(file_path, 3 + 19, 21)
    string += hover_req(file_path, 4 + 19, 17)
    string += hover_req(file_path, 5 + 19, 23)
    errcode, results = run_request(string, fortls_args=["-n", "1"])
    assert errcode == 0
    ref_results = [
        "```fortran90\nCHARACTER*17 :: A\n```",
        "```fortran90\nCHARACTER*17, DIMENSION(3,4) :: B\n```",
        "```fortran90\nCHARACTER*17, DIMENSION(9) :: V\n```",
        "```fortran90\nCHARACTER*(6+3) :: C\n```",
        "```fortran90\nCHARACTER*10, DIMENSION(3,4) :: D\n```",
        "```fortran90\nCHARACTER*(LEN(B)), DIMENSION(3,4) :: DD\n```",
    ]
    validate_hover(results, ref_results)


def test_var_name_asterisk():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "parse")})
    file_path = test_dir / "parse" / "test_kinds_and_dims.f90"
    string += hover_req(file_path, 26, 15)
    string += hover_req(file_path, 26, 22)
    string += hover_req(file_path, 26, 34)
    string += hover_req(file_path, 27, 15)
    string += hover_req(file_path, 28, 15)
    string += hover_req(file_path, 29, 15)
    string += hover_req(file_path, 31, 24)
    string += hover_req(file_path, 32, 32)
    # string += hover_req(file_path, 33, 32)  # FIXME: this is not displayed correctly
    errcode, results = run_request(string, fortls_args=["-n", "1"])
    assert errcode == 0
    ref_results = [
        "```fortran90\nCHARACTER*17 :: AA\n```",
        "```fortran90\nCHARACTER*17, DIMENSION(3,4) :: BB\n```",
        "```fortran90\nCHARACTER*17, DIMENSION(9) :: VV\n```",
        "```fortran90\nCHARACTER*(6+3) :: CC\n```",
        "```fortran90\nCHARACTER*(LEN(A)) :: AAA\n```",
        "```fortran90\nCHARACTER*10, DIMENSION(*) :: INPUT\n```",
        "```fortran90\nCHARACTER(LEN=200) :: F\n```",
        "```fortran90\nCHARACTER(KIND=4, LEN=200), DIMENSION(3,4) :: FF\n```",
        # "CHARACTER(KIND=4, LEN=100), DIMENSION(3,4)",
    ]
    validate_hover(results, ref_results)


def test_intent():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    file_path = test_dir / "hover" / "intent.f90"
    string += hover_req(file_path, 2, 31)
    string += hover_req(file_path, 3, 29)
    string += hover_req(file_path, 4, 34)
    string += hover_req(file_path, 5, 35)
    string += hover_req(file_path, 6, 35)
    errcode, results = run_request(string, fortls_args=["-n", "1"])
    assert errcode == 0
    ref_results = [
        """```fortran90\nINTEGER(4), INTENT(IN) :: arg1\n```""",
        """```fortran90\nINTEGER, INTENT(OUT) :: arg2\n```""",
        """```fortran90\nINTEGER(4), INTENT(INOUT) :: arg3\n```""",
        """```fortran90\nINTEGER(4), INTENT(IN OUT) :: arg4\n```""",
        """```fortran90\nREAL, OPTIONAL, INTENT(IN) :: arg5\n```""",
    ]
    validate_hover(results, ref_results)


def test_multiline_func_args():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "hover")})
    file_path = test_dir / "hover" / "functions.f90"
    string += hover_req(file_path, 58, 22)
    string += hover_req(file_path, 59, 22)
    string += hover_req(file_path, 60, 22)

    errcode, results = run_request(string, fortls_args=["-n", "1"])
    assert errcode == 0
    ref_results = [
        "```fortran90\nINTEGER, INTENT(IN) :: val1\n```",
        "```fortran90\nINTEGER, INTENT(IN) :: val2\n```",
        "```fortran90\nREAL :: val4\n```",
    ]
    validate_hover(results, ref_results)


def test_intrinsics():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "hover")})
    file_path = test_dir / "hover" / "functions.f90"
    string += hover_req(file_path, 39, 23)
    errcode, results = run_request(string, fortls_args=["-n", "1"])
    assert errcode == 0
    path = (
        test_dir.parent.parent
        / "fortls"
        / "parsers"
        / "internal"
        / "intrinsic.procedures.markdown.json"
    )
    with open(path, encoding="utf-8") as f:
        intrinsics = json.load(f)
    ref_results = ["\n-----\n" + intrinsics["SIZE"]]
    validate_hover(results, ref_results)


def test_types():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir / "hover")})
    file_path = test_dir / "hover" / "types.f90"
    string += hover_req(file_path, 3, 25)
    string += hover_req(file_path, 6, 44)
    string += hover_req(file_path, 9, 35)

    errcode, results = run_request(string, fortls_args=["-n", "1"])
    assert errcode == 0
    ref_results = [
        "```fortran90\nTYPE, ABSTRACT :: base_t\n```",
        "```fortran90\nTYPE, ABSTRACT, EXTENDS(base_t) :: extends_t\n```",
        "```fortran90\nTYPE, EXTENDS(extends_t) :: a_t\n```",
    ]
    validate_hover(results, ref_results)



from setup_tests import test_dir

from fortls.parsers.internal.parser import FortranFile


def test_line_continuations():
    file_path = test_dir / "parse" / "line_continuations.f90"
    file = FortranFile(str(file_path))
    err_str, _ = file.load_from_disk()
    assert err_str is None
    try:
        file.parse()
        assert True
    except Exception as e:
        print(e)
        assert False


def test_submodule():
    file_path = test_dir / "parse" / "submodule.f90"
    file = FortranFile(str(file_path))
    err_str, _ = file.load_from_disk()
    assert err_str is None
    try:
        ast = file.parse()
        assert True
        assert ast.scope_list[0].name == "val"
        assert ast.scope_list[0].ancestor_name == "p1"
        assert ast.scope_list[1].name == ""
        assert ast.scope_list[1].ancestor_name == "p2"
    except Exception as e:
        print(e)
        assert False


def test_private_visibility_interfaces():
    file_path = test_dir / "vis" / "private.f90"
    file = FortranFile(str(file_path))
    err_str, _ = file.load_from_disk()
    file.parse()
    assert err_str is None


def test_end_scopes_semicolon():
    file_path = test_dir / "parse" / "trailing_semicolon.f90"
    file = FortranFile(str(file_path))
    err_str, _ = file.load_from_disk()
    ast = file.parse()
    assert err_str is None
    assert not ast.end_errors



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import TYPE_CHECKING

from fortls.constants import ASSOC_TYPE_ID
from fortls.helper_functions import get_var_stack

from .block import Block
from .utilities import climb_type_tree, find_in_scope
from .variable import Variable

if TYPE_CHECKING:
    from .ast import FortranAST


@dataclass
class AssociateMap:
    var: Variable
    bind_name: str
    link_name: str


class Associate(Block):
    def __init__(self, file_ast: FortranAST, line_number: int, name: str):
        super().__init__(file_ast, line_number, name)
        self.links: list[AssociateMap] = []  # holds the info to associate variables

    def get_type(self, no_link=False):
        return ASSOC_TYPE_ID

    def get_desc(self):
        return "ASSOCIATE"

    def create_binding_variable(
        self, file_ast: FortranAST, line_number: int, bind_name: str, link_name: str
    ) -> Variable:
        """Create a new variable to be linked upon resolution to the real variable
        that contains the information of the mapping from the parent scope to the
        ASSOCIATE block scope.

        Parameters
        ----------
        file_ast : fortran_ast
            AST file
        line_number : int
            Line number
        bind_name : str
            Name of the ASSOCIATE block variable
        link_name : str
            Name of the parent scope variable

        Returns
        -------
        fortran_var
            Variable object holding the ASSOCIATE block variable, pending resolution
        """
        new_var = Variable(file_ast, line_number, bind_name, "UNKNOWN", [])
        self.links.append(AssociateMap(new_var, bind_name, link_name))
        return new_var

    def resolve_link(self, obj_tree):
        # Loop through the list of the associated variables map and resolve the links
        # find the AST node that that corresponds to the variable with link_name
        for assoc in self.links:
            # TODO: extract the dimensions component from the link_name
            # re.sub(r'\(.*\)', '', link_name) removes the dimensions component
            # keywords = re.match(r'(.*)\((.*)\)', link_name).groups()
            # now pass the keywords through the dimension_parser and set the keywords
            # in the associate object. Hover should now pick the local keywords
            # over the linked_object keywords
            assoc.link_name = re.sub(r"\(.*\)", "", assoc.link_name)
            var_stack = get_var_stack(assoc.link_name)
            is_member = len(var_stack) > 1
            if is_member:
                type_scope = climb_type_tree(var_stack, self, obj_tree)
                if type_scope is None:
                    continue
                var_obj = find_in_scope(type_scope, var_stack[-1], obj_tree)
            else:
                var_obj = find_in_scope(self, assoc.link_name, obj_tree)
            if var_obj is not None:
                assoc.var.link_obj = var_obj

    def require_link(self):
        return True



from __future__ import annotations

import pytest

from fortls.regex_patterns import create_src_file_exts_regex


@pytest.mark.parametrize(
    "input_exts, input_files, matches",
    [
        (
            [],
            [
                "test.f",
                "test.F",
                "test.f90",
                "test.F90",
                "test.f03",
                "test.F03",
                "test.f18",
                "test.F18",
                "test.f77",
                "test.F77",
                "test.f95",
                "test.F95",
                "test.for",
                "test.FOR",
                "test.fpp",
                "test.FPP",
            ],
            [True] * 16,
        ),
        ([], ["test.ff", "test.f901", "test.f90.ff"], [False, False, False]),
        ([r"\.inc"], ["test.inc", "testinc", "test.inc2"], [True, False, False]),
        (["inc.*"], ["test.inc", "testinc", "test.inc2"], [True, True, True]),
    ],
)
def test_src_file_exts(
    input_exts: list[str],
    input_files: list[str],
    matches: list[bool],
):
    regex = create_src_file_exts_regex(input_exts)
    results = [bool(regex.search(file)) for file in input_files]
    assert results == matches



from __future__ import annotations


# Helper classes
class Use:
    """AST node for USE statement"""

    def __init__(
        self,
        mod_name: str,
        only_list: set[str] = None,
        rename_map: dict[str, str] = None,
        line_number: int = 0,
    ):
        if only_list is None:
            only_list = set()
        if rename_map is None:
            rename_map = {}
        self.mod_name: str = mod_name.lower()
        self._line_no: int = line_number
        self.only_list: set[str] = only_list
        self.rename_map: dict[str, str] = rename_map
        if only_list:
            self.only_list: set[str] = {only.lower() for only in only_list}
        if rename_map:
            self.rename_map = {k.lower(): v.lower() for k, v in rename_map.items()}

    @property
    def line_number(self):
        return self._line_no

    @line_number.setter
    def line_number(self, line_number: int):
        self._line_no = line_number

    def rename(self, only_list: list[str] = None):
        """Rename ONLY:, statements"""
        if only_list is None:
            only_list = []
        if not only_list:
            only_list = self.only_list
        return [self.rename_map.get(only_name, only_name) for only_name in only_list]



from __future__ import annotations

import copy
from typing import TYPE_CHECKING
from typing import Type as T

from fortls.constants import (
    BLOCK_TYPE_ID,
    FUNCTION_TYPE_ID,
    INTERFACE_TYPE_ID,
    MODULE_TYPE_ID,
    SUBMODULE_TYPE_ID,
    SUBROUTINE_TYPE_ID,
)
from fortls.json_templates import range_json

from .base import FortranObj
from .diagnostics import Diagnostic
from .imports import Import
from .utilities import find_in_scope

if TYPE_CHECKING:
    from .ast import FortranAST
    from .use import Use


class Scope(FortranObj):
    def __init__(
        self,
        file_ast: FortranAST,
        line_number: int,
        name: str,
        keywords: list = None,
    ):
        super().__init__()
        if keywords is None:
            keywords = []
        self.file_ast: FortranAST = file_ast
        self.sline: int = line_number
        self.eline: int = line_number
        self.name: str = name
        self.children: list[T[Scope]] = []
        self.members: list = []
        self.use: list[Use | Import] = []
        self.keywords: list = keywords
        self.inherit = None
        self.parent = None
        self.contains_start = None
        self.implicit_line = None
        self.FQSN: str = self.name.lower()
        if file_ast.enc_scope_name is not None:
            self.FQSN = f"{file_ast.enc_scope_name.lower()}::{self.name.lower()}"

    def copy_from(self, copy_source: Scope):
        # Pass the reference, we don't want shallow copy since that would still
        # result into 2 versions of attributes between copy_source and self
        for k, v in copy_source.__dict__.items():
            setattr(self, k, v)

    def add_use(self, use_mod: Use | Import):
        self.use.append(use_mod)

    def set_inherit(self, inherit_type):
        self.inherit = inherit_type

    def set_parent(self, parent_obj):
        self.parent = parent_obj

    def set_implicit(self, implicit_flag, line_number):
        self.implicit_vars = implicit_flag
        self.implicit_line = line_number

    def mark_contains(self, line_number):
        if self.contains_start is not None:
            raise ValueError
        self.contains_start = line_number

    def add_child(self, child):
        self.children.append(child)
        child.set_parent(self)

    def update_fqsn(self, enc_scope=None):
        if enc_scope is not None:
            self.FQSN = f"{enc_scope.lower()}::{self.name.lower()}"
        else:
            self.FQSN = self.name.lower()
        for child in self.children:
            child.update_fqsn(self.FQSN)

    def add_member(self, member):
        self.members.append(member)

    def get_children(self, public_only=False) -> list[T[FortranObj]]:
        if not public_only:
            return copy.copy(self.children)
        pub_children = []
        for child in self.children:
            if (child.vis < 0) or ((self.def_vis < 0) and (child.vis <= 0)):
                continue
            if child.name.startswith("#GEN_INT"):
                pub_children.append(child)
                continue
            pub_children.append(child)
        return pub_children

    def check_definitions(self, obj_tree) -> list[Diagnostic]:
        """Check for definition errors in scope"""
        fqsn_dict: dict[str, int] = {}
        errors: list[Diagnostic] = []
        known_types: dict[str, FortranObj] = {}

        for child in self.children:
            # Skip masking/double checks for interfaces
            if child.get_type() == INTERFACE_TYPE_ID:
                continue
            # Check other variables in current scope
            if child.FQSN in fqsn_dict:
                if child.sline < fqsn_dict[child.FQSN]:
                    fqsn_dict[child.FQSN] = child.sline - 1
            else:
                fqsn_dict[child.FQSN] = child.sline - 1

        contains_line = -1
        if self.get_type() in (
            MODULE_TYPE_ID,
            SUBMODULE_TYPE_ID,
            SUBROUTINE_TYPE_ID,
            FUNCTION_TYPE_ID,
        ):
            contains_line = (
                self.contains_start if self.contains_start is not None else self.eline
            )
        # Detect interface definitions
        is_interface = (
            self.parent is not None
            and self.parent.get_type() == INTERFACE_TYPE_ID
            and not self.is_mod_scope()
        )

        for child in self.children:
            if child.name.startswith("#"):
                continue
            line_number = child.sline - 1
            # Check for type definition in scope
            def_error, known_types = child.check_definition(
                obj_tree, known_types=known_types, interface=is_interface
            )
            if def_error is not None:
                errors.append(def_error)
            # Detect contains errors
            if contains_line >= child.sline and child.get_type(no_link=True) in (
                SUBROUTINE_TYPE_ID,
                FUNCTION_TYPE_ID,
            ):
                new_diag = Diagnostic(
                    line_number,
                    message="Subroutine/Function definition before CONTAINS statement",
                    severity=1,
                )
                errors.append(new_diag)
            # Skip masking/double checks for interfaces and members
            if (
                self.get_type() == INTERFACE_TYPE_ID
                or child.get_type() == INTERFACE_TYPE_ID
            ):
                continue
            # Check other variables in current scope
            if child.FQSN in fqsn_dict and line_number > fqsn_dict[child.FQSN]:
                new_diag = Diagnostic(
                    line_number,
                    message=f'Variable "{child.name}" declared twice in scope',
                    severity=1,
                    find_word=child.name,
                )
                new_diag.add_related(
                    path=self.file_ast.path,
                    line=fqsn_dict[child.FQSN],
                    message="First declaration",
                )
                errors.append(new_diag)
                continue
            # Check for masking from parent scope in subroutines, functions, and blocks
            if self.parent is not None and self.get_type() in (
                SUBROUTINE_TYPE_ID,
                FUNCTION_TYPE_ID,
                BLOCK_TYPE_ID,
            ):
                parent_var = find_in_scope(self.parent, child.name, obj_tree)
                if parent_var is not None:
                    # Ignore if function return variable
                    if (
                        self.get_type() == FUNCTION_TYPE_ID
                        and parent_var.FQSN == self.FQSN
                    ):
                        continue

                    new_diag = Diagnostic(
                        line_number,
                        message=(
                            f'Variable "{child.name}" masks variable in parent scope'
                        ),
                        severity=2,
                        find_word=child.name,
                    )
                    new_diag.add_related(
                        path=parent_var.file_ast.path,
                        line=parent_var.sline - 1,
                        message="First declaration",
                    )
                    errors.append(new_diag)

        return errors

    def check_use(self, obj_tree):
        errors = []
        last_use_line = -1
        for use_stmnt in self.use:
            last_use_line = max(last_use_line, use_stmnt.line_number)
            if type(use_stmnt) is Import:
                if (self.parent is None) or (
                    self.parent.get_type() != INTERFACE_TYPE_ID
                ):
                    new_diag = Diagnostic(
                        use_stmnt.line_number - 1,
                        message="IMPORT statement outside of interface",
                        severity=1,
                    )
                    errors.append(new_diag)
                continue
            if use_stmnt.mod_name not in obj_tree:
                new_diag = Diagnostic(
                    use_stmnt.line_number - 1,
                    message=f'Module "{use_stmnt.mod_name}" not found in project',
                    severity=3,
                    find_word=use_stmnt.mod_name,
                )
                errors.append(new_diag)
        if (self.implicit_line is not None) and (last_use_line >= self.implicit_line):
            new_diag = Diagnostic(
                self.implicit_line - 1,
                message="USE statements after IMPLICIT statement",
                severity=1,
                find_word="IMPLICIT",
            )
            errors.append(new_diag)
        return errors

    def add_subroutine(self, interface_string, no_contains=False):
        edits = []
        line_number = self.eline - 1
        if (self.contains_start is None) and (not no_contains):
            first_sub_line = line_number
            for child in self.children:
                if child.get_type() in (SUBROUTINE_TYPE_ID, FUNCTION_TYPE_ID):
                    first_sub_line = min(first_sub_line, child.sline - 1)
            edits.append(
                {
                    **range_json(first_sub_line, 0, first_sub_line, 0),
                    "newText": "CONTAINS\n",
                }
            )
        edits.append(
            {
                **range_json(line_number, 0, line_number, 0),
                "newText": interface_string + "\n",
            }
        )
        return self.file_ast.path, edits



from __future__ import annotations

import os
import pprint

import json5

from .helper_functions import only_dirs, resolve_globs
from .jsonrpc import JSONRPC2Connection, ReadWriter, path_from_uri
from .langserver import LangServer
from .parsers.internal.parser import FortranFile


class DebugError(Exception):
    """Base class for debug CLI."""


class ParameterError(DebugError):
    """Exception raised for errors in the parameters."""


def is_debug_mode(args):
    debug_flags = [
        "debug_diagnostics",
        "debug_symbols",
        "debug_completion",
        "debug_signature",
        "debug_definition",
        "debug_hover",
        "debug_implementation",
        "debug_references",
        "debug_rename",
        "debug_actions",
        "debug_rootpath",
        "debug_workspace_symbols",
    ]
    return any(getattr(args, flag, False) for flag in debug_flags)


def debug_lsp(args, settings):
    debug_functions = {
        "debug_rootpath": debug_rootpath,
        "debug_diagnostics": debug_diagnostics,
        "debug_symbols": debug_symbols,
        "debug_workspace_symbols": debug_workspace_symbols,
        "debug_completion": debug_completion,
        "debug_hover": debug_hover,
        "debug_signature": debug_signature,
        "debug_definition": debug_definition,
        "debug_references": debug_references,
        "debug_implementation": debug_implementation,
        "debug_rename": debug_rename,
        "debug_actions": debug_actions,
    }

    r, w = os.pipe()
    with os.fdopen(r, "rb") as buffer_in, os.fdopen(w, "wb") as buffer_out:
        server = LangServer(
            conn=JSONRPC2Connection(ReadWriter(buffer_in, buffer_out)),
            settings=settings,
        )
        for flag, function in debug_functions.items():
            if getattr(args, flag, False):
                function(args, server)
                separator()


def debug_rootpath(args, server):
    if not os.path.isdir(args.debug_rootpath):
        raise DebugError("'debug_rootpath' not specified for debug request")
    print('\nTesting "initialize" request:')
    print(f'  Root = "{args.debug_rootpath}"')
    server.serve_initialize({"params": {"rootPath": args.debug_rootpath}})

    separator()
    if len(server.post_messages) == 0:
        print("  Successful!")
    else:
        print("  Successful with errors:")
        for message in server.post_messages:
            print(f"    {message[1]}")
    print("\n  Source directories:")
    for source_dir in server.source_dirs:
        print(f"    {source_dir}")


def debug_diagnostics(args, server):
    def lsp_request():
        server.serve_onSave({"params": {"textDocument": {"uri": args.debug_filepath}}})
        results, _ = server.get_diagnostics(args.debug_filepath)
        return results

    def format_results(results, _):
        sev_map = ["ERROR", "WARNING", "INFO"]
        if len(results) == 0:
            print("No errors or warnings")
        else:
            print("Reported Diagnostics:")
        for diag in results:
            sline = diag["range"]["start"]["line"]
            message = diag["message"]
            sev = sev_map[diag["severity"] - 1]
            print(f'  {sline:5d}:{sev}  "{message}"')

    debug_generic(
        args,
        "textDocument/publishDiagnostics",
        lsp_request,
        format_results,
        loc_needed=False,
    )


def debug_symbols(args, server):
    def lsp_request():
        server.serve_onSave({"params": {"textDocument": {"uri": args.debug_filepath}}})
        return server.serve_document_symbols(
            {"params": {"textDocument": {"uri": args.debug_filepath}}}
        )

    def format_results(results, _):
        for symbol in results:
            sline = symbol["location"]["range"]["start"]["line"]
            parent = "null"
            if "containerName" in symbol:
                parent = symbol["containerName"]
            print(
                f"  line {sline:5d}  symbol -> "
                f"{symbol['kind']:3d}:{symbol['name']:30} parent = {parent}"
            )

    debug_generic(
        args,
        "textDocument/documentSymbol",
        lsp_request,
        format_results,
        loc_needed=False,
    )


def debug_workspace_symbols(args, server):
    def lsp_request():
        if args.debug_rootpath is None:
            raise DebugError("'debug_rootpath' not specified for debug request")
        return server.serve_workspace_symbol(
            {"params": {"query": args.debug_workspace_symbols}}
        )

    def format_results(results, args):
        for symbol in results:
            path = path_from_uri(symbol["location"]["uri"])
            sline = symbol["location"]["range"]["start"]["line"]
            parent = "null"
            if "containerName" in symbol:
                parent = symbol["containerName"]
            print(
                f"  {parent}::{sline}  symbol -> {symbol['name']:30} parent = "
                f"{os.path.relpath(path, args.debug_rootpath)}"
            )

    debug_generic(
        args,
        "workspace/symbol",
        lsp_request,
        format_results,
        loc_needed=False,
    )


def debug_completion(args, server):
    def lsp_request():
        server.serve_onSave({"params": {"textDocument": {"uri": args.debug_filepath}}})
        return server.serve_autocomplete(
            {
                "params": {
                    "textDocument": {"uri": args.debug_filepath},
                    "position": {
                        "line": args.debug_line - 1,
                        "character": args.debug_char - 1,
                    },
                }
            }
        )

    def format_results(results, _):
        for obj in results:
            print(f"    {obj['kind']}: {obj['label']} -> {obj['detail']}")

    debug_generic(args, "textDocument/completion", lsp_request, format_results)


def debug_hover(args, server):
    def lsp_request():
        server.serve_onSave({"params": {"textDocument": {"uri": args.debug_filepath}}})
        return server.serve_hover(
            {
                "params": {
                    "textDocument": {"uri": args.debug_filepath},
                    "position": {
                        "line": args.debug_line - 1,
                        "character": args.debug_char - 1,
                    },
                }
            }
        )

    def format_results(results, _):
        contents = results["contents"]
        if isinstance(contents, dict):
            print(contents["value"])
        else:
            print(contents)

    debug_generic(args, "textDocument/hover", lsp_request, format_results)


def debug_signature(args, server):
    def lsp_request():
        server.serve_onSave({"params": {"textDocument": {"uri": args.debug_filepath}}})
        return server.serve_signature(
            {
                "params": {
                    "textDocument": {"uri": args.debug_filepath},
                    "position": {
                        "line": args.debug_line - 1,
                        "character": args.debug_char - 1,
                    },
                }
            }
        )

    def format_results(results, _):
        active_param = results.get("activeParameter", 0)
        print(f"    Active param = {active_param}")
        active_signature = results.get("activeSignature", 0)
        print(f"    Active sig   = {active_signature}")
        for i, signature in enumerate(results["signatures"]):
            print(f"    {signature['label']}")
            for j, obj in enumerate(signature["parameters"]):
                active_mark = " "
                if (i == active_signature) and (j == active_param):
                    active_mark = "*"
                arg_desc = obj.get("documentation")
                if arg_desc is not None:
                    print(f"{active_mark}     {arg_desc} :: {obj['label']}")
                else:
                    print(f"{active_mark}     {obj['label']}")

    debug_generic(args, "textDocument/signatureHelp", lsp_request, format_results)


def debug_definition(args, server):
    def lsp_request():
        server.serve_onSave({"params": {"textDocument": {"uri": args.debug_filepath}}})
        return server.serve_definition(
            {
                "params": {
                    "textDocument": {"uri": args.debug_filepath},
                    "position": {
                        "line": args.debug_line - 1,
                        "character": args.debug_char - 1,
                    },
                }
            }
        )

    def format_results(results, _):
        print(f'    URI  = "{results["uri"]}"')
        print(f'    Line = {results["range"]["start"]["line"] + 1}')
        print(f'    Char = {results["range"]["start"]["character"] + 1}')

    debug_generic(args, "textDocument/definition", lsp_request, format_results)


def debug_references(args, server):
    def lsp_request():
        server.serve_onSave({"params": {"textDocument": {"uri": args.debug_filepath}}})
        return server.serve_references(
            {
                "params": {
                    "textDocument": {"uri": args.debug_filepath},
                    "position": {
                        "line": args.debug_line - 1,
                        "character": args.debug_char - 1,
                    },
                }
            }
        )

    def format_results(results, _):
        for result in results:
            print(
                f"  {result['uri']}  ({result['range']['start']['line'] + 1}"
                f", {result['range']['start']['character'] + 1})"
            )

    debug_generic(args, "textDocument/references", lsp_request, format_results)


def debug_implementation(args, server):
    def lsp_request():
        server.serve_onSave({"params": {"textDocument": {"uri": args.debug_filepath}}})
        return server.serve_implementation(
            {
                "params": {
                    "textDocument": {"uri": args.debug_filepath},
                    "position": {
                        "line": args.debug_line - 1,
                        "character": args.debug_char - 1,
                    },
                }
            }
        )

    def format_results(results, _):
        print(f'    URI  = "{results["uri"]}"')
        print(f'    Line = {results["range"]["start"]["line"] + 1}')
        print(f'    Char = {results["range"]["start"]["character"] + 1}')

    debug_generic(args, "textDocument/implementation", lsp_request, format_results)


def debug_rename(args, server):
    def lsp_request():
        server.serve_onSave({"params": {"textDocument": {"uri": args.debug_filepath}}})
        return server.serve_rename(
            {
                "params": {
                    "textDocument": {"uri": args.debug_filepath},
                    "position": {
                        "line": args.debug_line - 1,
                        "character": args.debug_char - 1,
                    },
                    "newName": args.debug_rename,
                }
            }
        )

    def format_results(results, _):
        for uri, changes in results["changes"].items():
            path = path_from_uri(uri)
            file_obj = server.workspace.get(path)
            if file_obj is not None:
                file_contents = file_obj.contents_split
                process_file_changes(path, changes, file_contents)
            else:
                print(f'Unknown file: "{path}"')

    debug_generic(args, "textDocument/rename", lsp_request, format_results)


def debug_actions(args, server):
    def lsp_request():
        server.serve_onSave({"params": {"textDocument": {"uri": args.debug_filepath}}})
        return server.serve_codeActions(
            {
                "params": {
                    "textDocument": {"uri": args.debug_filepath},
                    "range": {
                        "start": {
                            "line": args.debug_line - 1,
                            "character": args.debug_char - 1,
                        },
                        "end": {
                            "line": args.debug_line - 1,
                            "character": args.debug_char - 1,
                        },
                    },
                }
            }
        )

    def process_results(results, _):
        pp = pprint.PrettyPrinter(indent=2, width=120)
        for result in results:
            print(f"Kind = '{result['kind']}', Title = '{result['title']}'")
            for edit_uri, edit_change in result["edit"]["changes"].items():
                print(f"\nChange: URI = '{edit_uri}'")
                pp.pprint(edit_change)
            print()

    debug_generic(args, "textDocument/getActions", lsp_request, process_results)


def process_file_changes(file_path, changes, file_contents):
    print(f'File: "{file_path}"')
    for change in changes:
        start_line = change["range"]["start"]["line"]
        end_line = change["range"]["end"]["line"]
        start_col = change["range"]["start"]["character"]
        end_col = change["range"]["end"]["character"]
        print(f"  {start_line + 1}, {end_line + 1}")
        new_contents = []
        for i in range(start_line, end_line + 1):
            line = file_contents[i]
            print(f"  - {line}")
            line_content = line
            if i == start_line:
                line_content = line[:start_col] + change["newText"]
            if i == end_line:
                line_content += line[end_col:]
            new_contents.append(line_content)
        for line in new_contents:
            print(f"  + {line}")
        print()


def debug_parser(args):
    """Debug the parser of the Language Server
    Triggered by `--debug_parser` option.

    Parameters
    ----------
    args : Namespace
        The arguments parsed from the `ArgumentParser`
    """

    def locate_config(root: str) -> str | None:
        default_conf_files = [args.config, ".fortlsrc", ".fortls.json5", ".fortls"]
        present_conf_files = [
            os.path.isfile(os.path.join(root, f)) for f in default_conf_files
        ]
        if not any(present_conf_files):
            return None

        # Load the first config file found
        for f, present in zip(default_conf_files, present_conf_files):
            if not present:
                continue
            config_path = os.path.join(root, f)
            return config_path

    def read_config(root: str | None):
        pp_suffixes = None
        pp_defs = {}
        include_dirs = set()
        if root is None:
            return pp_suffixes, pp_defs, include_dirs

        # Check for config files
        config_path = locate_config(root)
        print(f"  Config file = {config_path}")
        if config_path is None or not os.path.isfile(config_path):
            return pp_suffixes, pp_defs, include_dirs

        try:
            with open(config_path, encoding="utf-8") as fhandle:
                config_dict = json5.load(fhandle)
                pp_suffixes = config_dict.get("pp_suffixes", None)
                pp_defs = config_dict.get("pp_defs", {})
                for path in config_dict.get("include_dirs", set()):
                    include_dirs.update(only_dirs(resolve_globs(path, root)))

                if isinstance(pp_defs, list):
                    pp_defs = {key: "" for key in pp_defs}
        except ValueError as e:
            print(f"Error {e} while parsing '{config_path}' settings file")

        return pp_suffixes, pp_defs, include_dirs

    print("\nTesting parser")
    separator()

    ensure_file_accessible(args.debug_filepath)
    pp_suffixes, pp_defs, include_dirs = read_config(args.debug_rootpath)

    print(f'  File = "{args.debug_filepath}"')
    file_obj = FortranFile(args.debug_filepath, pp_suffixes)
    err_str, _ = file_obj.load_from_disk()
    if err_str:
        raise DebugError(f"Reading file failed: {err_str}")
    print(f"  Detected format: {'fixed' if file_obj.fixed else 'free'}")
    print("\n" + "=" * 80 + "\nParser Output\n" + "=" * 80 + "\n")
    file_ast = file_obj.parse(debug=True, pp_defs=pp_defs, include_dirs=include_dirs)
    print("\n" + "=" * 80 + "\nObject Tree\n" + "=" * 80 + "\n")
    for obj in file_ast.get_scopes():
        print(f"{obj.get_type()}: {obj.FQSN}")
        print_children(obj)
    print("\n" + "=" * 80 + "\nExportable Objects\n" + "=" * 80 + "\n")
    for _, obj in file_ast.global_dict.items():
        print(f"{obj.get_type()}: {obj.FQSN}")
    separator()


def ensure_file_accessible(filepath: str):
    """Ensure the file exists and is accessible, raising an error if not."""
    if not os.path.isfile(filepath):
        raise DebugError(f"File '{filepath}' does not exist or is not accessible")
    print(f'  File = "{filepath}"')


def check_request_params(args, loc_needed=True):
    ensure_file_accessible(args.debug_filepath)
    if loc_needed:
        if args.debug_line is None:
            raise ParameterError("'debug_line' not specified for debug request")
        print(f"  Line = {args.debug_line}")
        if args.debug_char is None:
            raise ParameterError("'debug_char' not specified for debug request")
        print(f"  Char = {args.debug_char}\n")


def debug_generic(args, test_label, lsp_request, format_results, loc_needed=True):
    print(f'\nTesting "{test_label}" request:')
    check_request_params(args, loc_needed)
    results = lsp_request()
    separator()
    print_results(results, format_results, args)


def print_results(results, format_results, args):
    """Helper function to print results based on detail level requested."""
    if results is None:
        print("    No result found!")
        return

    if args.debug_full_result:
        print(json5.dumps(results, indent=2))
        return

    format_results(results, args)


def print_children(obj, indent=""):
    for child in obj.get_children():
        print(f"  {indent}{child.get_type()}: {child.FQSN}")
        print_children(child, indent + "  ")


def separator():
    print("=" * 80)



from __future__ import annotations

import os
import re
from re import Pattern

from fortls.ftypes import IncludeInfo
from fortls.json_templates import diagnostic_json

from .diagnostics import Diagnostic
from .imports import Import, ImportTypes
from .interface import Interface
from .program import Program
from .scope import Scope
from .use import Use
from .variable import Variable


class FortranAST:
    def __init__(self, file_obj=None):
        self.file = file_obj
        self.path: str | None = file_obj.path if file_obj is not None else None
        self.global_dict: dict = {}
        self.scope_list: list = []
        self.variable_list: list = []
        self.public_list: list = []
        self.private_list: list = []
        self.scope_stack: list = []
        self.end_stack: list = []
        self.pp_if: list = []
        self.include_statements: list = []
        self.end_errors: list = []
        self.parse_errors: list = []
        self.inherit_objs: list = []
        self.linkable_objs: list = []
        self.external_objs: list = []
        self.none_scope = None
        self.inc_scope = None
        self.current_scope = None
        self.end_scope_regex: Pattern | None = None
        self.enc_scope_name: str | None = None
        self.last_obj = None
        self.pending_doc: str | None = None

    def create_none_scope(self):
        """Create empty scope to hold non-module contained items"""
        if self.none_scope is not None:
            raise ValueError
        self.none_scope = Program(self, 1, "main")
        self.add_scope(
            self.none_scope, re.compile(r"[ ]*END[ ]*PROGRAM", re.I), exportable=False
        )

    def get_enc_scope_name(self):
        """Get current enclosing scope name"""
        return None if self.current_scope is None else self.current_scope.FQSN

    def add_scope(
        self,
        new_scope: Scope,
        end_scope_regex: Pattern[str],
        exportable: bool = True,
        req_container: bool = False,
    ):
        self.scope_list.append(new_scope)
        if new_scope.require_inherit():
            self.inherit_objs.append(new_scope)
        if new_scope.require_link():
            self.linkable_objs.append(new_scope)
        if self.current_scope is None:
            if req_container:
                self.create_none_scope()
                new_scope.FQSN = f"{self.none_scope.FQSN}::{new_scope.name.lower()}"
                self.current_scope.add_child(new_scope)
                self.scope_stack.append(self.current_scope)
            elif exportable:
                self.global_dict[new_scope.FQSN] = new_scope
        else:
            self.current_scope.add_child(new_scope)
            self.scope_stack.append(self.current_scope)
        if self.end_scope_regex is not None:
            self.end_stack.append(self.end_scope_regex)
        self.current_scope = new_scope
        self.end_scope_regex = end_scope_regex
        self.enc_scope_name = self.get_enc_scope_name()
        self.last_obj = new_scope
        if self.pending_doc is not None:
            self.last_obj.add_doc(self.pending_doc)
            self.pending_doc = None

    def end_scope(self, line_number: int, check: bool = True):
        if (
            (self.current_scope is None) or (self.current_scope is self.none_scope)
        ) and check:
            self.end_errors.append([-1, line_number])
            return
        self.current_scope.end(line_number)
        if len(self.scope_stack) > 0:
            self.current_scope = self.scope_stack.pop()
        else:
            self.current_scope = None
        if len(self.end_stack) > 0:
            self.end_scope_regex = self.end_stack.pop()
        else:
            self.end_scope_regex = None
        self.enc_scope_name = self.get_enc_scope_name()

    def add_variable(self, new_var: Variable):
        if self.current_scope is None:
            self.create_none_scope()
            new_var.FQSN = f"{self.none_scope.FQSN}::{new_var.name.lower()}"
        self.current_scope.add_child(new_var)
        self.variable_list.append(new_var)
        if new_var.is_external:
            self.external_objs.append(new_var)
        if new_var.require_link():
            self.linkable_objs.append(new_var)
        self.last_obj = new_var
        if self.pending_doc is not None:
            self.last_obj.add_doc(self.pending_doc)
            self.pending_doc = None

    def add_int_member(self, key):
        self.current_scope.add_member(key)

    def add_private(self, name: str):
        self.private_list.append(f"{self.enc_scope_name}::{name}")

    def add_public(self, name: str):
        self.public_list.append(f"{self.enc_scope_name}::{name}")

    def add_use(self, use_mod: Use | Import):
        if self.current_scope is None:
            self.create_none_scope()
        self.current_scope.add_use(use_mod)

    def add_include(self, path: str, line_number: int):
        self.include_statements.append(IncludeInfo(line_number, path, None, []))

    def add_doc(self, doc_string: str, forward: bool = False):
        if not doc_string:
            return
        if forward:
            self.pending_doc = doc_string
        elif self.last_obj is not None:
            self.last_obj.add_doc(doc_string)

    def add_error(self, msg: str, sev: int, ln: int, sch: int, ech: int = None):
        """Add a Diagnostic error, encountered during parsing, for a range
        in the document.

        Parameters
        ----------
        msg : str
            Error message
        sev : int
            Severity, Error, Warning, Notification
        ln : int
            Line number
        sch : int
            Start character
        ech : int
            End character
        """
        # Convert from Editor line numbers 1-base index to LSP index which is 0-based
        self.parse_errors.append(diagnostic_json(ln - 1, sch, ln - 1, ech, msg, sev))

    def start_ppif(self, line_number: int):
        self.pp_if.append([line_number - 1, -1])

    def end_ppif(self, line_number: int):
        if len(self.pp_if) > 0:
            self.pp_if[-1][1] = line_number - 1

    def get_scopes(self, line_number: int | None = None):
        """Get a list of all the scopes present in the line number provided.

        Parameters
        ----------
        line_number : int, optional
            Document line number, if None return all document scopes, by default None

        Returns
        -------
        Variable,Type,Function,Subroutine,Module,Program,Interface,BlockData
            A list of scopes
        """
        if line_number is None:
            return self.scope_list
        scope_list = []
        for scope in self.scope_list:
            if not scope.sline <= line_number <= scope.eline:
                continue
            if type(scope.parent) is Interface:
                for use_stmnt in scope.use:
                    if type(use_stmnt) is not Import:
                        continue
                    # Exclude the parent and all other scopes
                    if use_stmnt.import_type == ImportTypes.NONE:
                        return [scope]
            scope_list.append(scope)
            scope_list.extend(iter(scope.get_ancestors()))
        if scope_list or self.none_scope is None:
            return scope_list
        return [self.none_scope]

    def get_inner_scope(self, line_number: int):
        scope_sline = -1
        curr_scope = None
        for scope in self.scope_list:
            if scope.sline > scope_sline and scope.sline <= line_number <= scope.eline:
                curr_scope = scope
                scope_sline = scope.sline
        if (curr_scope is None) and (self.none_scope is not None):
            return self.none_scope
        return curr_scope

    def get_object(self, FQSN: str):
        def find_child_by_name(parent, name):
            for child in parent.children:
                if child.name == name:
                    return child
                if child.name.startswith("#GEN_INT"):
                    found = next(
                        (
                            int_child
                            for int_child in child.get_children()
                            if int_child.name == name
                        ),
                        None,
                    )
                    if found:
                        return found
            return None

        parts = FQSN.split("::")
        current = self.global_dict.get(parts[0])

        # Look for non-exportable scopes
        if current is None:
            current = next(
                (scope for scope in self.scope_list if scope.FQSN == parts[0]), None
            )
            if current is None:
                return None

        for part in parts[1:]:
            current = find_child_by_name(current, part)
            if current is None:
                return None

        return current

    def resolve_includes(self, workspace, path: str | None = None):
        file_dir = os.path.dirname(self.path)
        for inc in self.include_statements:
            file_path = os.path.normpath(os.path.join(file_dir, inc.path))
            if path and path != file_path:
                continue
            parent_scope = self.get_inner_scope(inc.line_number)
            added_entities = inc.scope_objs
            if file_path in workspace:
                include_file = workspace[file_path]
                include_ast = include_file.ast
                inc.file = include_file
                if include_ast.none_scope:
                    if include_ast.inc_scope is None:
                        include_ast.inc_scope = include_ast.none_scope
                    # Remove old objects
                    for obj in added_entities:
                        parent_scope.children.remove(obj)
                    added_entities = []
                    for child in include_ast.inc_scope.children:
                        added_entities.append(child)
                        if parent_scope is not None:
                            parent_scope.add_child(child)
                            child.update_fqsn(parent_scope.FQSN)
                    include_ast.none_scope = parent_scope
                    inc.scope_objs = added_entities

    def resolve_links(self, obj_tree, link_version):
        for inherit_obj in self.inherit_objs:
            inherit_obj.resolve_inherit(obj_tree, inherit_version=link_version)
        for linkable_obj in self.linkable_objs:
            linkable_obj.resolve_link(obj_tree)

    def close_file(self, line_number: int):
        # Close open scopes
        while self.current_scope is not None:
            self.end_scope(line_number, check=False)
        # Close and delist none_scope
        if self.none_scope is not None:
            self.none_scope.end(line_number)
            self.scope_list.remove(self.none_scope)
        # Tasks to be done when file parsing is finished
        for private_name in self.private_list:
            obj = self.get_object(private_name)
            if obj is not None:
                obj.set_visibility(-1)
        for public_name in self.public_list:
            obj = self.get_object(public_name)
            if obj is not None:
                obj.set_visibility(1)

    def check_file(self, obj_tree):
        errors = []
        tmp_list = self.scope_list[:]  # shallow copy
        if self.none_scope is not None:
            tmp_list += [self.none_scope]
        for error in self.end_errors:
            if error[0] >= 0:
                message = f"Unexpected end of scope at line {error[0]}"
            else:
                message = "Unexpected end statement: No open scopes"
            errors.append(Diagnostic(error[1] - 1, message=message, severity=1))
        for scope in tmp_list:
            if not scope.check_valid_parent():
                errors.append(
                    Diagnostic(
                        scope.sline - 1,
                        message=f'Invalid parent for "{scope.get_desc()}" declaration',
                        severity=1,
                    )
                )
            errors += scope.check_use(obj_tree)
            errors += scope.check_definitions(obj_tree)
            errors += scope.get_diagnostics()
        return errors, self.parse_errors



from __future__ import annotations

from fortls.json_templates import diagnostic_json, location_json
from fortls.jsonrpc import path_to_uri


class Diagnostic:
    def __init__(
        self, sline: int, message: str, severity: int = 1, find_word: str = None
    ):
        self.sline: int = sline
        self.message: str = message
        self.severity: int = severity
        self.find_word: str = find_word
        self.has_related: bool = False
        self.related_path = None
        self.related_line = None
        self.related_message = None

    def add_related(self, path: str, line: int, message: str):
        self.has_related = True
        self.related_path = path
        self.related_line = line
        self.related_message = message

    def build(self, file_obj):
        schar = echar = 0
        if self.find_word is not None:
            self.sline, obj_range = file_obj.find_word_in_code_line(
                self.sline, self.find_word
            )
            if obj_range.start >= 0:
                schar = obj_range.start
                echar = obj_range.end
        diag = diagnostic_json(
            self.sline, schar, self.sline, echar, self.message, self.severity
        )
        if self.has_related:
            diag["relatedInformation"] = [
                {
                    **location_json(
                        path_to_uri(self.related_path), self.related_line, 0
                    ),
                    "message": self.related_message,
                }
            ]
        return diag



from __future__ import annotations

from typing import TYPE_CHECKING

from .use import Use

if TYPE_CHECKING:
    from .module import Module
    from .scope import Scope


class ImportTypes:
    DEFAULT = -1
    NONE = 0
    ALL = 1
    ONLY = 2


class Import(Use):
    """AST node for IMPORT statement"""

    def __init__(
        self,
        name: str,
        import_type: ImportTypes = ImportTypes.DEFAULT,
        only_list: set[str] = None,
        rename_map: dict[str, str] = None,
        line_number: int = 0,
    ):
        if only_list is None:
            only_list = set()
        if rename_map is None:
            rename_map = {}
        super().__init__(name, only_list, rename_map, line_number)
        self.import_type = import_type
        self._scope: Scope | Module | None = None

    @property
    def scope(self):
        """Parent scope of IMPORT statement i.e. parent of the interface"""
        return self._scope

    @scope.setter
    def scope(self, scope: Scope):
        self._scope = scope



from __future__ import annotations

from dataclasses import dataclass
from re import I, compile
from typing import Pattern


@dataclass(frozen=True)
class FortranRegularExpressions:
    USE: Pattern = compile(
        r"[ ]*USE([, ]+(?:INTRINSIC|NON_INTRINSIC))?[ :]+(\w*)([, ]+ONLY[ :]+)?",
        I,
    )
    IMPORT: Pattern = compile(
        r"[ ]*IMPORT"
        r"(?:"
        r"[ ]*,[ ]*(?P<spec>ALL|NONE)"  # import, [all | none]
        r"|"  # or
        r"[ ]*,[ ]*(?P<only>ONLY)[ ]*:[ ]*(?P<start1>[\w_])"  # import, only: name-list
        r"|"  # or
        r"[ ]+(?:::[ ]*)?(?P<start2>[\w_])"  # import [[::] name-list]
        r")?",  # standalone import
        I,
    )
    INCLUDE: Pattern = compile(r"[ ]*INCLUDE[ :]*[\'\"]([^\'\"]*)", I)
    CONTAINS: Pattern = compile(r"[ ]*(CONTAINS)[ ]*$", I)
    IMPLICIT: Pattern = compile(r"[ ]*IMPLICIT[ ]+([a-z]*)", I)
    #: Parse procedure keywords but not if they start with , or ( or end with , or )
    #: This is to avoid parsing as keywords variables named pure, impure, etc.
    SUB_MOD: Pattern = compile(
        r"[ ]*(?!<[,\()][ ]*)\b(PURE|IMPURE|ELEMENTAL|RECURSIVE)\b(?![,\)][ ]*)", I
    )
    SUB: Pattern = compile(r"[ ]*SUBROUTINE[ ]+(\w+)", I)
    END_SUB: Pattern = compile(r"SUBROUTINE", I)
    FUN: Pattern = compile(r"[ ]*FUNCTION[ ]+(\w+)", I)
    RESULT: Pattern = compile(r"RESULT[ ]*\((\w*)\)", I)
    END_FUN: Pattern = compile(r"FUNCTION", I)
    MOD: Pattern = compile(r"[ ]*MODULE[ ]+(\w+)", I)
    END_MOD: Pattern = compile(r"MODULE", I)
    SUBMOD: Pattern = compile(r"[ ]*SUBMODULE[ ]*\(", I)
    END_SMOD: Pattern = compile(r"SUBMODULE", I)
    END_PRO: Pattern = compile(r"(MODULE)?[ ]*PROCEDURE", I)
    BLOCK: Pattern = compile(r"[ ]*([a-z_]\w*[ ]*:[ ]*)?BLOCK|CRITICAL(?!\w)", I)
    END_BLOCK: Pattern = compile(r"BLOCK|CRITICAL", I)
    DO: Pattern = compile(r"[ ]*(?:[a-z_]\w*[ ]*:[ ]*)?DO([ ]+[0-9]*|$)", I)
    END_DO: Pattern = compile(r"DO", I)
    WHERE: Pattern = compile(r"[ ]*WHERE[ ]*\(", I)
    END_WHERE: Pattern = compile(r"WHERE", I)
    IF: Pattern = compile(r"[ ]*(?:[a-z_]\w*[ ]*:[ ]*)?IF[ ]*\(", I)
    THEN: Pattern = compile(r"\)[ ]*THEN$", I)
    END_IF: Pattern = compile(r"IF", I)
    ASSOCIATE: Pattern = compile(r"[ ]*ASSOCIATE[ ]*\(", I)
    END_ASSOCIATE: Pattern = compile(r"ASSOCIATE", I)
    END_FIXED: Pattern = compile(r"[ ]*([0-9]*)[ ]*CONTINUE", I)
    SELECT: Pattern = compile(
        r"[ ]*(?:[a-z_]\w*[ ]*:[ ]*)?SELECT[ ]*" r"(CASE|TYPE)[ ]*\(([\w=> ]*)",
        I,
    )
    SELECT_TYPE: Pattern = compile(r"[ ]*(TYPE|CLASS)[ ]+IS[ ]*\(([\w ]*)", I)
    SELECT_DEFAULT: Pattern = compile(r"[ ]*CLASS[ ]+DEFAULT", I)
    END_SELECT: Pattern = compile(r"SELECT", I)
    PROG: Pattern = compile(r"[ ]*PROGRAM[ ]+(\w+)", I)
    END_PROG: Pattern = compile(r"PROGRAM", I)
    INT: Pattern = compile(r"[ ]*(ABSTRACT)?[ ]*INTERFACE[ ]*(\w*)", I)
    END_INT: Pattern = compile(r"INTERFACE", I)
    END_WORD: Pattern = compile(
        r"[ ]*END[ ]*(DO|WHERE|IF|BLOCK|CRITICAL|ASSOCIATE|SELECT"
        r"|TYPE|ENUM|MODULE|SUBMODULE|PROGRAM|INTERFACE"
        r"|SUBROUTINE|FUNCTION|PROCEDURE|FORALL)?([ ]+(?!\W)|$)",
        I,
    )
    TYPE_DEF: Pattern = compile(r"[ ]*(TYPE)[, :]+", I)
    EXTENDS: Pattern = compile(r"EXTENDS[ ]*\((\w*)\)", I)
    GENERIC_PRO: Pattern = compile(
        r"[ ]*(GENERIC)[, ]*(PRIVATE|PUBLIC)?[ ]*::[ ]*[a-z]", I
    )
    GEN_ASSIGN: Pattern = compile(r"(ASSIGNMENT|OPERATOR)\(", I)
    END_TYPED: Pattern = compile(r"TYPE", I)
    ENUM_DEF: Pattern = compile(r"[ ]*ENUM[, ]+", I)
    END_ENUMD: Pattern = compile(r"ENUM", I)
    VAR: Pattern = compile(
        r"[ ]*(INTEGER|REAL|DOUBLE[ ]*PRECISION|COMPLEX"
        r"|DOUBLE[ ]*COMPLEX|CHARACTER|LOGICAL|PROCEDURE"
        r"|EXTERNAL|CLASS|TYPE)",  # external :: variable is handled by this
        I,
    )
    KIND_SPEC: Pattern = compile(r"[ ]*([*]?\([ ]*[\w*:]|\*[ ]*[0-9:]*)", I)
    KEYWORD_LIST: Pattern = compile(
        r"[ ]*,[ ]*(PUBLIC|PRIVATE|ALLOCATABLE|"
        r"POINTER|TARGET|DIMENSION[ ]*\(|"
        r"OPTIONAL|INTENT[ ]*\([ ]*(?:IN|OUT|IN[ ]*OUT)[ ]*\)|DEFERRED|NOPASS|"
        r"PASS[ ]*\(\w*\)|SAVE|PARAMETER|EXTERNAL|"
        r"CONTIGUOUS)",
        I,
    )
    PARAMETER_VAL: Pattern = compile(r"\w*[\s\&]*=(([\s\&]*[\w\.\-\+\*\/\'\"])*)", I)
    TATTR_LIST: Pattern = compile(
        r"[ ]*,[ ]*(PUBLIC|PRIVATE|ABSTRACT|EXTENDS\(\w*\))", I
    )
    VIS: Pattern = compile(r"[ ]*\b(PUBLIC|PRIVATE)\b", I)
    WORD: Pattern = compile(r"[a-z_][\w\$]*", I)
    NUMBER: Pattern = compile(
        r"[\+\-]?(\b\d+\.?\d*|\.\d+)(_\w+|d[\+\-]?\d+|e[\+\-]?\d+(_\w+)?)?(?!\w)",
        I,
    )
    LOGICAL: Pattern = compile(r".true.|.false.", I)
    SUB_PAREN: Pattern = compile(r"\([\w, ]*\)", I)
    # KIND_SPEC_MATCH: Pattern = compile(r"\([\w, =*]*\)", I)

    SQ_STRING: Pattern = compile(r"\'[^\']*\'", I)
    DQ_STRING: Pattern = compile(r"\"[^\"]*\"", I)
    LINE_LABEL: Pattern = compile(r"[ ]*([0-9]+)[ ]+", I)
    NON_DEF: Pattern = compile(r"[ ]*(CALL[ ]+[a-z_]|[a-z_][\w%]*[ ]*=)", I)
    # Fixed format matching rules
    FIXED_COMMENT: Pattern = compile(r"([!cd*])", I)
    FIXED_CONT: Pattern = compile(r"( {5}[\S])")
    FIXED_DOC: Pattern = compile(r"(?:[!cd\*])([<>!])", I)
    FIXED_OPENMP: Pattern = compile(r"[!c\*]\$OMP", I)
    # Free format matching rules
    FREE_COMMENT: Pattern = compile(r"([ ]*!)")
    FREE_CONT: Pattern = compile(r"([ ]*&)")
    FREE_DOC: Pattern = compile(r"[ ]*!([<>!])")
    FREE_OPENMP: Pattern = compile(r"[ ]*!\$OMP", I)
    FREE_FORMAT_TEST: Pattern = compile(r"[ ]{1,4}[a-z]", I)
    # Preprocessor matching rules
    DEFINED: Pattern = compile(r"defined[ ]*\(?[ ]*([a-z_]\w*)[ ]*\)?", I)
    PP_REGEX: Pattern = compile(r"[ ]*#[ ]*(if |ifdef|ifndef|else|elif|endif)", I)
    PP_DEF: Pattern = compile(
        r"[ ]*#[ ]*(define|undef|undefined)[ ]*(\w+)(\([ ]*([ \w,]*?)[ ]*\))?",
        I,
    )
    PP_DEF_TEST: Pattern = compile(r"(![ ]*)?defined[ ]*\([ ]*(\w*)[ ]*\)$", I)
    PP_INCLUDE: Pattern = compile(r"[ ]*#[ ]*include[ ]*([\"\w\.]*)", I)
    PP_ANY: Pattern = compile(r"^[ ]*#:?[ ]*(\w+)")
    # Context matching rules
    CALL: Pattern = compile(r"[ ]*CALL[ ]+[\w%]*$", I)
    INT_STMNT: Pattern = compile(r"^[ ]*[a-z]*$", I)
    TYPE_STMNT: Pattern = compile(r"[ ]*(TYPE|CLASS)[ ]*(IS)?[ ]*$", I)
    PROCEDURE_STMNT: Pattern = compile(r"[ ]*(PROCEDURE)[ ]*$", I)
    PRO_LINK: Pattern = compile(r"[ ]*(MODULE[ ]*PROCEDURE )", I)
    SCOPE_DEF: Pattern = compile(
        r"[ ]*(MODULE|PROGRAM|SUBROUTINE|FUNCTION|INTERFACE)[ ]+", I
    )
    END: Pattern = compile(
        r"[ ]*(END)("
        r" |MODULE|PROGRAM|SUBROUTINE|FUNCTION|PROCEDURE|TYPE|DO|IF|SELECT)?",
        I,
    )
    # Object regex patterns
    CLASS_VAR: Pattern = compile(r"(TYPE|CLASS)[ ]*\(", I)
    DEF_KIND: Pattern = compile(r"(\w*)[ ]*\((?:KIND|LEN)?[ =]*(\w*)", I)
    OBJBREAK: Pattern = compile(r"[\/\-(.,+*<>=: ]", I)


# TODO: use this in the main code
def create_src_file_exts_regex(input_exts: list[str] = []) -> Pattern[str]:
    r"""Create a REGEX for which sources the Language Server should parse.

    Default extensions are (case insensitive):
    F F03 F05 F08 F18 F77 F90 F95 FOR FPP

    Parameters
    ----------
    input_exts : list[str], optional
        Additional list of file extensions to parse, in Python REGEX format
        that means special characters must be escaped
        , by default []

    Examples
    --------
    >>> regex = create_src_file_exts_regex([r"\.fypp", r"\.inc"])
    >>> regex.search("test.fypp")
    <re.Match object; span=(4, 9), match='.fypp'>
    >>> regex.search("test.inc")
    <re.Match object; span=(4, 8), match='.inc'>

    >>> regex = create_src_file_exts_regex([r"\.inc.*"])
    >>> regex.search("test.inc.1")
    <re.Match object; span=(4, 10), match='.inc.1'>

    Invalid regex expressions will cause the function to revert to the default
    extensions

    >>> regex = create_src_file_exts_regex(["*.inc"])
    >>> regex.search("test.inc") is None
    True

    Returns
    -------
    Pattern[str]
        A compiled regular expression for matching file extensions
    """
    import re

    DEFAULT = r"\.[fF](77|90|95|03|05|08|18|[oO][rR]|[pP]{2})?"
    EXPRESSIONS = [DEFAULT]
    try:
        EXPRESSIONS.extend(input_exts)
        # Add its expression as an OR and force they match the end of the string
        return re.compile(rf"(({'$)|('.join(EXPRESSIONS)}$))")
    except re.error:
        # TODO: Add a warning to the logger
        return re.compile(rf"({DEFAULT}$)")


def create_src_file_exts_str(input_exts: list[str] = []) -> Pattern[str]:
    """This is a version of create_src_file_exts_regex that takes a list
    sanitises the list of input_exts before compiling the regex.
    For more info see create_src_file_exts_regex
    """
    import re

    input_exts = [re.escape(ext) for ext in input_exts]
    return create_src_file_exts_regex(input_exts)



from __future__ import annotations

from setup_tests import run_request, test_dir, write_rpc_request


def test_hover():
    def hover_req(file_path: str, ln: int, col: int) -> str:
        return write_rpc_request(
            1,
            "textDocument/hover",
            {
                "textDocument": {"uri": str(file_path)},
                "position": {"line": ln, "character": col},
            },
        )

    def check_return(result_array, checks):
        assert len(result_array) == len(checks)
        for i, check in enumerate(checks):
            assert result_array[i]["contents"]["value"] == check

    root_dir = test_dir / "pp"
    string = write_rpc_request(1, "initialize", {"rootPath": str(root_dir)})
    file_path = root_dir / "preproc.F90"
    string += hover_req(file_path, 5, 8)  # user defined type
    string += hover_req(file_path, 7, 30)  # variable
    string += hover_req(file_path, 7, 40)  # multi-lin variable
    string += hover_req(file_path, 8, 7)  # function with if conditional
    string += hover_req(file_path, 9, 7)  # multiline function with if conditional
    string += hover_req(file_path, 10, 15)  # defined without ()
    file_path = root_dir / "preproc_keywords.F90"
    string += hover_req(file_path, 6, 2)  # ignores PP across Fortran line continuations
    file_path = root_dir / "preproc_else.F90"
    string += hover_req(file_path, 8, 12)
    string += hover_req(file_path, 18, 12)
    file_path = root_dir / "preproc_elif.F90"
    string += hover_req(file_path, 22, 15)
    string += hover_req(file_path, 24, 10)
    file_path = root_dir / "preproc_elif_elif_skip.F90"
    string += hover_req(file_path, 30, 23)
    file_path = root_dir / "preproc_if_elif_else.F90"
    string += hover_req(file_path, 30, 23)
    file_path = root_dir / "preproc_if_elif_skip.F90"
    string += hover_req(file_path, 30, 23)
    config = str(root_dir / ".pp_conf.json")
    errcode, results = run_request(string, ["--config", config])
    assert errcode == 0

    # Reference solution
    ref_results = (
        "```fortran90\n#define PCType character*(80)\n```",
        "```fortran90\n#define PETSC_ERR_INT_OVERFLOW 84\n```",
        "```fortran90\n#define varVar 55\n```",
        (
            "```fortran90\n#define ewrite(priority, format)"
            " if (priority <= 3) write((priority), format)\n```"
        ),
        (
            "```fortran90\n#define ewrite2(priority, format)"
            " if (priority <= 3) write((priority), format)\n```"
        ),
        "```fortran90\n#define SUCCESS .true.\n```",
        "```fortran90\nREAL, CONTIGUOUS, POINTER, DIMENSION(:) :: var1\n```",
        "```fortran90\nINTEGER :: var0\n```",
        "```fortran90\nREAL :: var1\n```",
        "```fortran90\nINTEGER :: var2\n```",
        "```fortran90\nINTEGER, INTENT(INOUT) :: var\n```",
        "```fortran90\nINTEGER, PARAMETER :: res = 0+1+0+0\n```",
        "```fortran90\nINTEGER, PARAMETER :: res = 0+0+0+1\n```",
        "```fortran90\nINTEGER, PARAMETER :: res = 1+0+0+0\n```",
    )
    assert len(ref_results) == len(results) - 1
    check_return(results[1:], ref_results)



from __future__ import annotations

from typing import TYPE_CHECKING

from fortls.constants import FUNCTION_TYPE_ID
from fortls.helper_functions import get_keywords

from .subroutine import Subroutine

if TYPE_CHECKING:
    from .ast import FortranAST
    from .variable import Variable


class Function(Subroutine):
    def __init__(
        self,
        file_ast: FortranAST,
        line_number: int,
        name: str,
        args: str = "",
        mod_flag: bool = False,
        keywords: list = None,
        keyword_info: dict = None,
        result_type: str = None,
        result_name: str = None,
    ):
        super().__init__(file_ast, line_number, name, args, mod_flag, keywords)
        self.args: str = args.replace(" ", "").lower()
        self.args_snip: str = self.args
        self.arg_objs: list = []
        self.in_children: list = []
        self.missing_args: list = []
        self.mod_scope: bool = mod_flag
        self.result_name: str = result_name
        self.result_type: str = result_type
        self.result_obj: Variable = None
        self.keyword_info: dict = keyword_info
        # Set the implicit result() name to be the function name
        if self.result_name is None:
            self.result_name = self.name
        # Used in Associated blocks
        if self.keyword_info is None:
            self.keyword_info = {}

    def copy_interface(self, copy_source: Function):
        # Call the parent class method
        child_names = super().copy_interface(copy_source)
        # Return specific options
        self.result_name = copy_source.result_name
        self.result_type = copy_source.result_type
        self.result_obj = copy_source.result_obj
        if (
            copy_source.result_obj is not None
            and copy_source.result_obj.name.lower() not in child_names
        ):
            self.in_children.append(copy_source.result_obj)

    def resolve_link(self, obj_tree):
        self.resolve_arg_link(obj_tree)
        result_var_lower = self.result_name.lower()
        for child in self.children:
            if child.name.lower() == result_var_lower:
                self.result_obj = child
                # Update result value and type
                self.result_name = child.name
                self.result_type = child.get_desc()

    def get_type(self, no_link=False):
        return FUNCTION_TYPE_ID

    def get_desc(self):
        token = "FUNCTION"
        return f"{self.result_type} {token}" if self.result_type else token

    def is_callable(self):
        return False

    def get_hover(self, long: bool = False, drop_arg: int = -1) -> tuple[str, str]:
        """Construct the hover message for a FUNCTION.
        Two forms are produced here the `long` i.e. the normal for hover requests

        [MODIFIERS] FUNCTION NAME([ARGS]) RESULT(RESULT_VAR)
          TYPE, [ARG_MODIFIERS] :: [ARGS]
          TYPE, [RESULT_MODIFIERS] :: RESULT_VAR

        note: intrinsic functions will display slightly different,
        `RESULT_VAR` and its `TYPE` might not always be present

        short form, used when functions are arguments in functions and subroutines:

        FUNCTION NAME([ARGS]) :: ARG_LIST_NAME

        Parameters
        ----------
        long : bool, optional
            toggle between long and short hover results, by default False
        drop_arg : int, optional
            Ignore argument at position `drop_arg` in the argument list, by default -1

        Returns
        -------
        tuple[str, bool]
            String representative of the hover message and the `long` flag used
        """
        fun_sig, _ = self.get_snippet(drop_arg=drop_arg)
        # short hover messages do not include the result()
        fun_sig += f" RESULT({self.result_name})" if long else ""
        keyword_list = get_keywords(self.keywords)
        keyword_list.append("FUNCTION")

        hover_array = [f"{' '.join(keyword_list)} {fun_sig}"]
        hover_array, docs = self.get_docs_full(hover_array, long, drop_arg)
        # Only append the return value if using long form
        if self.result_obj and long:
            # Parse the documentation from the result variable
            arg_doc, doc_str = self.result_obj.get_hover()
            if doc_str is not None:
                docs.append(f"\n**Return:**  \n`{self.result_obj.name}`{doc_str}")
            hover_array.append(arg_doc)
        # intrinsic functions, where the return type is missing but can be inferred
        elif self.result_type and long:
            # prepend type to function signature
            hover_array[0] = f"{self.result_type} {hover_array[0]}"
        return "\n ".join(hover_array), "  \n".join(docs)

    # TODO: fix this
    def get_interface(self, name_replace=None, drop_arg=-1, change_strings=None):
        fun_sig, _ = self.get_snippet(name_replace=name_replace)
        fun_sig += f" RESULT({self.result_name})"
        # XXX:
        keyword_list = []
        if self.result_type:
            keyword_list.append(self.result_type)
        keyword_list += get_keywords(self.keywords)
        keyword_list.append("FUNCTION ")

        interface_array = self.get_interface_array(
            keyword_list, fun_sig, drop_arg, change_strings
        )
        if self.result_obj is not None:
            arg_doc, docs = self.result_obj.get_hover()
            interface_array.append(f"{arg_doc} :: {self.result_obj.name}")
        name = name_replace if name_replace is not None else self.name
        interface_array.append(f"END FUNCTION {name}")
        return "\n".join(interface_array)



from __future__ import annotations

from typing import TYPE_CHECKING

from fortls.constants import SELECT_TYPE_ID

from .block import Block
from .variable import Variable

if TYPE_CHECKING:
    from .ast import FortranAST


class Select(Block):
    def __init__(
        self,
        file_ast: FortranAST,
        line_number: int,
        name: str,
        select_info,
    ):
        super().__init__(file_ast, line_number, name)
        self.select_type = select_info.type
        self.binding_name = None
        self.bound_var = None
        self.binding_type = None
        if self.select_type == 2:
            binding_split = select_info.binding.split("=>")
            if len(binding_split) == 1:
                self.bound_var = binding_split[0].strip()
            elif len(binding_split) == 2:
                self.binding_name = binding_split[0].strip()
                self.bound_var = binding_split[1].strip()
        elif self.select_type == 3:
            self.binding_type = select_info.binding
        # Close previous "TYPE IS" region if open
        if (
            (file_ast.current_scope is not None)
            and (file_ast.current_scope.get_type() == SELECT_TYPE_ID)
            and file_ast.current_scope.is_type_region()
        ):
            file_ast.end_scope(line_number)

    def get_type(self, no_link=False):
        return SELECT_TYPE_ID

    def get_desc(self):
        return "SELECT"

    def is_type_binding(self):
        return self.select_type == 2

    def is_type_region(self):
        return self.select_type in [3, 4]

    def create_binding_variable(self, file_ast, line_number, var_desc, case_type):
        if self.parent.get_type() != SELECT_TYPE_ID:
            return None
        binding_name = None
        bound_var = None
        if (self.parent is not None) and self.parent.is_type_binding():
            binding_name = self.parent.binding_name
            bound_var = self.parent.bound_var
        # Check for default case
        if (binding_name is not None) and (case_type != 4):
            bound_var = None
        # Create variable
        if binding_name is not None:
            return Variable(
                file_ast, line_number, binding_name, var_desc, [], link_obj=bound_var
            )
        elif bound_var is not None:
            return Variable(file_ast, line_number, bound_var, var_desc, [])
        return None



from setup_tests import run_request, test_dir, write_rpc_request


def check_msg(ref, res):
    assert ref["type"] == res["type"]
    assert ref["message"] == res["message"]


# def test_config_file_non_existent():
#     string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
#     errcode, results = run_request(string, ["-c", "fake.json"])
#
#     ref = {"type": 1, "message": "Configuration file 'fake.json' not found"}
#     assert errcode == 0
#     check_msg(ref, results[0])


def test_config_file_non_existent_options():
    string = write_rpc_request(1, "initialize", {"rootPath": str(test_dir)})
    errcode, results = run_request(string, ["-c", "wrong_syntax.json"])

    ref = {
        "type": 1,
        "message": (
            'Error: "<string>:2 Unexpected "," at column 18" while reading'
            ' "wrong_syntax.json" Configuration file'
        ),
    }
    assert errcode == 0
    check_msg(ref, results[0])



from __future__ import annotations

from typing import TYPE_CHECKING

from fortls.constants import INTERFACE_TYPE_ID

from .scope import Scope
from .utilities import find_in_scope

if TYPE_CHECKING:
    from .ast import FortranAST


class Interface(Scope):
    def __init__(
        self,
        file_ast: FortranAST,
        line_number: int,
        name: str,
        abstract: bool = False,
    ):
        super().__init__(file_ast, line_number, name)
        self.mems = []
        self.abstract = abstract
        self.external = name.startswith("#GEN_INT") and (not abstract)

    def get_type(self, no_link=False):
        return INTERFACE_TYPE_ID

    def get_desc(self):
        return "INTERFACE"

    def is_callable(self):
        return True

    def is_external_int(self):
        return self.external

    def is_abstract(self):
        return self.abstract

    def resolve_link(self, obj_tree):
        if self.parent is None:
            return
        self.mems = []
        for member in self.members:
            mem_obj = find_in_scope(self.parent, member, obj_tree)
            if mem_obj is not None:
                self.mems.append(mem_obj)

    def require_link(self):
        return True



from __future__ import annotations

import json
import pathlib

from pydantic import Field, create_model

from fortls.interface import cli


def create_schema(root: pathlib.Path | None = None):
    if not root:
        root = pathlib.Path(__file__).parent

    parser = cli("fortls")
    only_vals = {}
    for arg in parser._actions:
        if (
            arg.dest == "help"
            or arg.dest == "version"
            or arg.help == "==SUPPRESS=="
            or (arg.dest.startswith("debug") and arg.dest != "debug_log")
        ):
            continue
        val = arg.default
        desc: str = arg.help.replace("%(default)s", str(val))  # type: ignore
        only_vals[arg.dest] = (type(val), Field(val, description=desc))  # type: ignore

    m = create_model("fortls schema", **only_vals)
    m.__doc__ = "Schema for the fortls Fortran Language Server"

    with open(str(root / "fortls.schema.json"), "w", encoding="utf-8") as f:
        print(json.dumps(m.model_json_schema(), indent=2), file=f)
    print(f"Created schema file: {root / 'fortls.schema.json'}")


if __name__ == "__main__":
    create_schema()
